=== modified file 'lib/lp/bugs/configure.zcml'
--- lib/lp/bugs/configure.zcml	2011-07-18 19:18:29 +0000
+++ lib/lp/bugs/configure.zcml	2011-07-27 03:58:17 +0000
@@ -254,7 +254,9 @@
                     transitionToStatus
                     transitionToAssignee
                     transitionToMilestone
-                    transitionToTarget"/>
+                    transitionToTarget
+                    validateTransitionToTarget
+                    "/>
 
             <!-- XXX Brad Bollenbach 2006-09-29 bug=63000:
                  The permission checking for milestone and importance

=== modified file 'lib/lp/bugs/interfaces/bugtask.py'
--- lib/lp/bugs/interfaces/bugtask.py	2011-07-21 07:12:51 +0000
+++ lib/lp/bugs/interfaces/bugtask.py	2011-07-27 03:58:17 +0000
@@ -794,6 +794,12 @@
         value is set to None, date_assigned is also set to None.
         """
 
+    def validateTransitionToTarget(target):
+        """Check whether a transition to this target is legal.
+
+        :raises IllegalTarget: if the new target is not allowed.
+        """
+
     @mutator_for(target)
     @operation_parameters(
         target=copy_field(target))

=== modified file 'lib/lp/bugs/model/bugtask.py'
--- lib/lp/bugs/model/bugtask.py	2011-07-27 03:44:07 +0000
+++ lib/lp/bugs/model/bugtask.py	2011-07-27 04:34:19 +0000
@@ -1138,20 +1138,8 @@
             get_property_cache(self.bug)._known_viewers = set(
                 [self.assignee.id])
 
-    def transitionToTarget(self, target):
-        """See `IBugTask`.
-
-        This method allows changing the target of some bug
-        tasks. The rules it follows are similar to the ones
-        enforced implicitly by the code in
-        lib/canonical/launchpad/browser/bugtask.py#BugTaskEditView.
-        """
-
-        if self.target == target:
-            return
-
-        target_before_change = self.target
-
+    def validateTransitionToTarget(self, target):
+        """See `IBugTask`."""
         # Check if any series are involved. You can't retarget series
         # tasks. Except for DistroSeries/SourcePackage tasks, which can
         # only be retargetted to another SourcePackage in the same
@@ -1178,6 +1166,29 @@
 
         validate_target(self.bug, target)
 
+    def transitionToTarget(self, target):
+        """See `IBugTask`.
+
+        This method allows changing the target of some bug
+        tasks. The rules it follows are similar to the ones
+        enforced implicitly by the code in
+        lib/canonical/launchpad/browser/bugtask.py#BugTaskEditView.
+        """
+
+        if self.target == target:
+            return
+
+        self.validateTransitionToTarget(target)
+
+        target_before_change = self.target
+
+        if (self.milestone is not None and
+            self.milestone.target != target):
+            # If the milestone for this bugtask is set, we
+            # have to make sure that it's a milestone of the
+            # current target, or reset it to None
+            self.milestone = None
+
         # Inhibit validate_target_attribute, as we can't set them all
         # atomically, but we know the final result is correct.
         self._inhibit_target_check = True
@@ -1186,13 +1197,6 @@
         self._inhibit_target_check = False
         self.updateTargetNameCache()
 
-        if (self.milestone is not None and
-            self.milestone.target != target):
-            # If the milestone for this bugtask is set, we
-            # have to make sure that it's a milestone of the
-            # current target, or reset it to None
-            self.milestone = None
-
         # After the target has changed, we need to recalculate the maximum bug
         # heat for the new and old targets.
         if self.target != target_before_change:

=== modified file 'lib/lp/bugs/model/tests/test_bugtask.py'
--- lib/lp/bugs/model/tests/test_bugtask.py	2011-07-27 03:44:07 +0000
+++ lib/lp/bugs/model/tests/test_bugtask.py	2011-07-27 05:10:54 +0000
@@ -75,6 +75,7 @@
     TestCaseWithFactory,
     )
 from lp.testing.factory import LaunchpadObjectFactory
+from lp.testing.fakemethod import FakeMethod
 from lp.testing.matchers import HasQueryCount
 
 
@@ -1650,31 +1651,25 @@
                 self.assertEqual(0, len(recorder.events))
 
 
-class TestTransitionToTarget(TestCaseWithFactory):
-    """Tests for BugTask.transitionToTarget."""
+class TestValidateTransitionToTarget(TestCaseWithFactory):
+    """Tests for BugTask.validateTransitionToTarget."""
 
     layer = DatabaseFunctionalLayer
 
-    def makeAndTransition(self, old, new):
+    def makeAndCheckTransition(self, old, new):
         task = self.factory.makeBugTask(target=old)
-        p = self.factory.makePerson()
-        self.assertEqual(old, task.target)
-        old_state = Snapshot(task, providing=providedBy(task))
         with person_logged_in(task.owner):
-            task.bug.subscribe(p, p)
-            task.transitionToTarget(new)
-            notify(ObjectModifiedEvent(task, old_state, ["target"]))
-        return task
+            task.validateTransitionToTarget(new)
 
     def assertTransitionWorks(self, a, b):
         """Check that a transition between two targets works both ways."""
-        self.assertEqual(b, self.makeAndTransition(a, b).target)
-        self.assertEqual(a, self.makeAndTransition(b, a).target)
+        self.makeAndCheckTransition(a, b)
+        self.makeAndCheckTransition(b, a)
 
     def assertTransitionForbidden(self, a, b):
         """Check that a transition between two targets fails both ways."""
-        self.assertRaises(IllegalTarget, self.makeAndTransition, a, b)
-        self.assertRaises(IllegalTarget, self.makeAndTransition, b, a)
+        self.assertRaises(IllegalTarget, self.makeAndCheckTransition, a, b)
+        self.assertRaises(IllegalTarget, self.makeAndCheckTransition, b, a)
 
     def test_product_to_product_works(self):
         self.assertTransitionWorks(
@@ -1748,10 +1743,6 @@
         sp2 = self.factory.makeSourcePackage(distroseries=ds2)
         self.assertTransitionForbidden(sp1, sp2)
 
-    def test_transition_to_same_is_noop(self):
-        p = self.factory.makeProduct()
-        self.assertTransitionWorks(p, p)
-
     def test_validate_target_is_called(self):
         p = self.factory.makeProduct()
         task1 = self.factory.makeBugTask(target=p)
@@ -1763,16 +1754,90 @@
                 "A fix for this bug has already been requested for %s"
                 % p.displayname, task2.transitionToTarget, p)
 
+
+class TestTransitionToTarget(TestCaseWithFactory):
+    """Tests for BugTask.transitionToTarget."""
+
+    layer = DatabaseFunctionalLayer
+
+    def makeAndTransition(self, old, new):
+        task = self.factory.makeBugTask(target=old)
+        p = self.factory.makePerson()
+        self.assertEqual(old, task.target)
+        old_state = Snapshot(task, providing=providedBy(task))
+        with person_logged_in(task.owner):
+            task.bug.subscribe(p, p)
+            task.transitionToTarget(new)
+            notify(ObjectModifiedEvent(task, old_state, ["target"]))
+        return task
+
+    def assertTransitionWorks(self, a, b):
+        """Check that a transition between two targets works both ways."""
+        self.assertEqual(b, self.makeAndTransition(a, b).target)
+        self.assertEqual(a, self.makeAndTransition(b, a).target)
+
+    def test_transition_works(self):
+        self.assertTransitionWorks(
+            self.factory.makeProduct(),
+            self.factory.makeProduct())
+
+    def test_target_type_transition_works(self):
+        # A transition from one type of target to another works.
+        self.assertTransitionWorks(
+            self.factory.makeProduct(),
+            self.factory.makeDistributionSourcePackage())
+
+    def test_validation(self):
+        # validateTransitionToTarget is called before any transition.
+        p = self.factory.makeProduct()
+        task = self.factory.makeBugTask(target=p)
+
+        # Patch out validateTransitionToTarget to raise an exception
+        # that we can check. Also check that the target was not changed.
+        msg = self.factory.getUniqueString()
+        removeSecurityProxy(task).validateTransitionToTarget = FakeMethod(
+            failure=IllegalTarget(msg))
+        with person_logged_in(task.owner):
+            self.assertRaisesWithContent(
+                IllegalTarget, msg,
+                task.transitionToTarget, self.factory.makeProduct())
+        self.assertEqual(p, task.target)
+
+    def test_transition_to_same_is_noop(self):
+        # While a no-op transition would normally be rejected due to
+        # task duplication, transitionToTarget short-circuits.
+        p = self.factory.makeProduct()
+        self.assertTransitionWorks(p, p)
+
+    def test_milestone_unset_on_transition(self):
+        # A task's milestone is reset when its target changes.
+        product = self.factory.makeProduct()
+        task = self.factory.makeBugTask(target=product)
+        with person_logged_in(task.owner):
+            task.milestone = self.factory.makeMilestone(product=product)
+            task.transitionToTarget(self.factory.makeProduct())
+        self.assertIs(None, task.milestone)
+
     def test_milestone_preserved_if_transition_rejected(self):
-        series = self.factory.makeProductSeries()
-        task = self.factory.makeBugTask(target=series.product)
-        milestone = self.factory.makeMilestone(product=series.product)
+        # If validation rejects a transition, the milestone is not unset.
+        product = self.factory.makeProduct()
+        task = self.factory.makeBugTask(target=product)
         with person_logged_in(task.owner):
-            task.milestone = milestone
+            task.milestone = milestone = self.factory.makeMilestone(
+                product=product)
             self.assertRaises(
                 IllegalTarget,
                 task.transitionToTarget, self.factory.makeSourcePackage())
-            self.assertEqual(milestone, task.milestone)
+        self.assertEqual(milestone, task.milestone)
+
+    def test_targetnamecache_updated(self):
+        new_product = self.factory.makeProduct()
+        task = self.factory.makeBugTask()
+        with person_logged_in(task.owner):
+            task.transitionToTarget(new_product)
+        self.assertEqual(
+            new_product.bugtargetdisplayname,
+            removeSecurityProxy(task).targetnamecache)
 
 
 class TestBugTargetKeys(TestCaseWithFactory):

